import isNil from 'lodash/isNil';
import defaults from 'lodash/defaults';
import toPairs from 'lodash/toPairs';
import mapKeys from 'lodash/mapKeys';
import isUndefined from 'lodash/isUndefined';
import { fetch } from './polyfill';
import { checkStatus, normParams, parseObject } from './http-helper';
import middleware from './middleware';
import { split } from './filter';

const parseFilters = filtersIn => (filtersIn || []).filter(({ filter }) => !!filter).map(({ key, filter }) => [
  split(key, false).map(value => `f-${value}`).join('|'),
  filter,
]);

const makeQueryParams = ({ pst, psz, filters = [], sortBys = [], sortTypes = [], params = {} }) => {
  let stBy = sortBys.join(',');
  stBy = stBy || undefined;
  let stType = sortTypes.join(',');
  stType = stType || undefined;
  return [
    ...toPairs({ pst, psz, stBy, stType }),
    ...parseFilters(filters),
    ...normParams(mapKeys(params, (v, k) => `p-${k}`)),
  ].filter(v => v && !isUndefined((v[1])));
};

export const makeParams = (otherParams, { filters = [], sortBys = [], sortTypes = [], params = {} }) => {
  return [
    ...toPairs(otherParams),
    ...makeQueryParams({ filters, sortBys, sortTypes, params }),
  ].filter(v => v && !isUndefined((v[1])));
};

const defaultOptions = {
  headers: { Accept: 'application/json' },
};

/**
 * Requests a URL, returning a promise.
 *
 * @param  {string} url       The URL we want to request
 * @param auth
 * @param params
 * @param  {object} [options] The options we want to pass to "fetch"
 * @return {object}           An object containing either "data" or "err"
 */
export default async function request(url, params = {}, options = {}, auth = true) {
  const normedParams = normParams(params);
  const res = await middleware.get.onRequest(url, normedParams, options, auth);
  const newUrl = res ? res.url : url;
  const newParams = res ? res.params : normedParams;
  const newOptions = res ? res.options : options;
  let queryParams = newParams.map(([k, v]) => (isNil(v) ? k : `${k}=${encodeURIComponent(v)}`));
  queryParams = queryParams.join('&');
  let realUrl = newUrl;
  if (queryParams) {
    realUrl = `${url}?${queryParams}`;
  }
  return fetch(realUrl, defaults(newOptions, defaultOptions))
    .then(checkStatus)
    .then(resp => parseObject(resp, middleware.get.onResponse));
}

